"""
Ledger-related blueprints.
Functions for things like adding new events and productions.
"""

from flask import Blueprint, request, jsonify, make_response, json
from asimov import ledger

events_bp = Blueprint("events", __name__, url_prefix="/events")

@events_bp.route("/", methods=["GET"])
def show_ledger():
    """Return the entire contents of the ledger.

    Examples
    --------
    .. http:example:: curl wget python-requests

       GET /events HTTP/1.1
       Content-Type: application/json

    """

    events = ledger.events
    return make_response(jsonify(events), 200)

@events_bp.route("/<event>", methods=["GET"])
def show_event(event):
    """Return the entire contents of the ledger for a single event.

    Examples
    --------
    .. http:example:: curl wget python-requests

       GET /events/<event> HTTP/1.1

    """

    event = ledger.get_event(event=event)
    return make_response(event.to_dict(), 200)

@events_bp.route("/<event>", methods=["POST"])
def create_event(event):
    """Add a new event to the ledger.

    Examples
    --------
    .. http:example:: curl wget python-requests

       POST /events/<event> HTTP/1.1
       Content-Type: application/json

    """
    from asimov.event import Event
    new_event = Event(name=event, **request.get_json())
    ledger.add_event(new_event)
    return make_response(new_event.to_dict(), 201)

@events_bp.route("/<event>/productions", methods=['GET'])
def show_productions(event):
    """Return all of the productions for a given event.

    Examples
    --------
    .. http:example:: curl wget python-requests

       GET /events/<event>/productions HTTP/1.1
       Content-Type: application/json

    """

    filters = request.args

    event = ledger.get_event(event=event)
    return make_response(
        jsonify([production.to_dict() for production in event.productions]),
        200)





import os
import sys
import collections
import json

from math import floor

import click

# Ignore warnings from the condor module
import warnings
warnings.filterwarnings("ignore", module="htcondor")

from asimov import logging
from asimov import current_ledger as ledger
from asimov import condor
import asimov.pipelines

# Replace this with a better logfile handling module please
#from glob import glob
import asimov

from git.exc import GitCommandError

# Import CLI bits from elsewhere
from asimov import cli
from asimov.cli import configuration
from asimov.cli import report
from asimov.cli import monitor
from asimov.cli import review
from asimov.cli import manage
from asimov.cli import event
from asimov.cli import project
from asimov.cli import production
from asimov.cli import application


@click.version_option(asimov.__version__)
@click.group()
@click.pass_context
def olivaw(ctx):
    """
    This is the main olivaw program which runs the DAGs for each event issue.
    """

    # Check that we're running in an actual asimov project
    if not os.path.exists("asimov.conf") and ctx.invoked_subcommand != "init":
        # This isn't the root of an asimov project, let's fail.
        click.secho("This isn't an asimov project", fg="white", bg="red")
        sys.exit(1)
    pass

# Project initialisation
olivaw.add_command(project.init)
olivaw.add_command(project.clone)

olivaw.add_command(event.event)

# Building and submission
olivaw.add_command(manage.manage)
# Reporting commands
olivaw.add_command(report.report)
# Configuration commands
olivaw.add_command(configuration.configuration)
# Monitoring commands
olivaw.add_command(monitor.start)
olivaw.add_command(monitor.stop)
olivaw.add_command(monitor.monitor)
olivaw.add_command(event.event)
olivaw.add_command(production.production)
# Review commands
olivaw.add_command(review.review)
olivaw.add_command(application.apply)



"""
Supervisor: A package for interacting with 
long-running data analysis jobs.
"""


__author__ = """Daniel Williams"""
__email__ = 'daniel.williams@ligo.org'
__packagename__ = __name__

from pkg_resources import get_distribution, DistributionNotFound
from pkg_resources import resource_string
import os

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "dev"
    pass

try:
    import ConfigParser as configparser
except ImportError:
    import configparser
default_config = resource_string(__name__, '{}.conf'.format(__packagename__))

config = configparser.ConfigParser()
#if not config_file:

config.read_string(default_config.decode("utf8"))
config_locations = [os.path.join(os.curdir, "{}.conf".format(__packagename__)),
                    os.path.join(os.path.expanduser("~"),
                                 ".config", __packagename__, "{}.conf".format(__packagename__)),
                    os.path.join(os.path.expanduser("~"),
                                 ".{}".format(__packagename__)),
                    "/etc/{}".format(__packagename__)]

config_locations.reverse()

config.read([conffile for conffile in config_locations])



try:

    if config.get("general", "logger") == "file":
        from .logging import AsimovLogger
        logger = AsimovLogger(logfile="asimov.log")
    elif config.get("general", "logger") == "database":
        from .logging import DatabaseLogger
        logger = DatabaseLogger()
except configparser.NoOptionError:
    from .logging import AsimovLogger
    logger = AsimovLogger(logfile="asimov.log")

try:
    if config.get("ledger", "engine") == "gitlab":
        from .gitlab import GitlabLedger
        current_ledger = GitlabLedger()
    elif config.get("ledger", "engine") == "yamlfile":
        from .ledger import YAMLLedger
        current_ledger = YAMLLedger(config.get("ledger", "location"))
    else:
        current_ledger = None
except FileNotFoundError:
    logger.error("Could not find a valid ledger file.")
    current_ledger = None 



"""
Endpoints for productions.

Asimov enables a number of interactions with analyses (productions)
it is monitoring to be made via the web API.

These are:

+ Production creation
+ Production deletion
+ Updating unstarted productions
+ Changing the run status of a production
"""

from flask import Blueprint, request, jsonify, make_response, json
from asimov import ledger
from asimov.event import Production, DescriptionException

prods_bp = Blueprint("productions", __name__, url_prefix="/productions")


@prods_bp.route("/", methods=['GET', 'POST'])
@prods_bp.route("/<event>", methods=['GET', 'POST'])
def show_event_productions(event=None):
    """Return all of the productions in the ledger for a given event.

    .. http:get:: /productions

       :statuscode 200: Successful
    
       

       Example
       -------

       .. http:example:: curl wget python-requests

          GET /productions/ HTTP/1.1
          Content-Type: application/json


          HTTP/1.1 200 OK
          Content-Type: application/json

          [
             {
               "calibration": {},
               "comment": null,
               "event": "GW150914",
               "event time": 678,
               "pipeline": "bayeswave",
               "review": [],
               "status": "ready",
               "working directory": "working/GW150914"
             },
             {
                 "approximant": "SEOBNRv4PHM",
                 "calibration": {
                   "H1": "C01_offline/calibration/H1.dat",
                   "L1": "C01_offline/calibration/L1.dat",
                   "V1": "C01_offline/calibration/V1.dat"
                 },
                 "combination": {
                   "job": 231253218,
                   "productions": []
                 },
                 "comment": "RIFT job",
                 "data": {
                   "channels": {
                     "H1": "H1:DCS-CALIB_STRAIN_CLEAN_SUB60HZ_C01",
                     "L1": "L1:DCS-CALIB_STRAIN_CLEAN_SUB60HZ_C01",
                     "V1": "V1:Hrec_hoft_16384Hz"
                   },
                   "frame-types": {
                     "H1": "H1_HOFT_CLEAN_SUB60HZ_C01",
                     "L1": "L1_HOFT_CLEAN_SUB60HZ_C01",
                     "V1": "V1Online"
                   }
                 },
                 ...
               }
             ]

    .. http:get:: /productions/(event_id)

       Fetch the events for a single event ID.

       .. http:example:: curl wget python-requests

          GET /productions/GW150914 HTTP/1.1
          Content-Type: application/json


          HTTP/1.1 200 OK
          Content-Type: application/json

          [
            {
              "calibration": {},
              "comment": null,
              "event": "GW150914",
              "event time": 678,
              "pipeline": "bayeswave",
              "review": [],
              "status": "ready",
              "working directory": "working/GW150914"
            },
            {
              "approximant": "IMRPhenomXPHM",
              "calibration": {},
              "comment": "A test production",
              "event": "GW150914",
              "event time": 678,
              "pipeline": "bilby",
              "review": [],
              "status": "ready",
              "working directory": "working/GW150914"
            }
          ]

       :statuscode 200: Successful

    .. http:post:: /productions/(event_id)

       .. http:example:: curl wget python-requests

          POST /productions/GW170817 HTTP/1.1
          Content-Type: application/json

          {"approximant": "IMRPhenomXPHM",
           "comment": "A test production",
           "status": "ready",
           "pipeline": "bilby"}

       .. http:example:: curl wget python-requests

          POST /productions/ HTTP/1.1
          Content-Type: application/json

          {"approximant": "IMRPhenomXPHM",
           "event": "GW170817",
           "comment": "A test production",
           "status": "ready",
           "pipeline": "bilby"}

    """
    if request.method == "POST":
        return add_event_production(event)
    elif request.method == "GET":
        productions = ledger.get_productions(event=event, filters=request.args.to_dict())
        
    return make_response(
        jsonify([production.to_dict()[production.name] for production in productions]),
        200)


def add_event_production(event=None):
    """Add a new production to an event.
    """
    data = request.get_json()
    print("data", data)

    if not event:
        event = data['event']
    
    event = ledger.get_event(event)
    event_prods = event.productions
    names = [production.name for production in event_prods]
    
    #if "family" in data:
    #    family = data.pop("family")
    #else:
    #    family = "Prod"
    #family_entries = [int(name.split(family)[1]) for name in names if family in name]
        
    #
    print("GET JSON")
    production = request.get_json()
    print("GOT JSON")
    
    #if len(family_entries)>0:
    #    number = max(family_entries)+1
    #else:
    #    number = 0
    #production_dict = {f"{family}{number}": production}
    #print("dict", production_dict)

    production_dict = {"Test0": production}

    print("GET DICT")
    production = Production.from_dict(parameters=production_dict, event=event, issue=None)
    print("END")
    #
    try:
        ledger.add_production(event, production)
        response = make_response(jsonify(production.to_dict()), 201)
    except DescriptionException as e:
        response = make_response(jsonify({"_message": e}), 400)
    
    return response





from contextlib import contextmanager
from pathlib import Path
from copy import deepcopy
from asimov import logger
import os
import glob
import numpy as np
import collections

@contextmanager
def set_directory(path: (Path, str)):
    """
    Change to a different directory for the duration of the context.

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """

    origin = Path().absolute()
    try:
        #print(f"{origin} → {path}")
        logger.info(f"Working temporarily in {path}")
        os.chdir(path)
        yield
    finally:
        #print(f"{path} → {origin}")
        os.chdir(origin)


def find_calibrations(time):
    """
    Find the calibration file for a given time.
    """
    if time < 1190000000:
        dir = "/home/cal/public_html/uncertainty/O2C02"
        virgo = "/home/carl-johan.haster/projects/O2/C02_reruns/V_calibrationUncertaintyEnvelope_magnitude5p1percent_phase40mraddeg20microsecond.txt"
    elif time < 1290000000:
        dir = "/home/cal/public_html/uncertainty/O3C01"
        virgo = "/home/cbc/pe/O3/calibrationenvelopes/Virgo/V_O3a_calibrationUncertaintyEnvelope_magnitude5percent_phase35milliradians10microseconds.txt"
    data_llo = glob.glob(f"{dir}/L1/*LLO*FinalResults.txt")
    times_llo = {int(datum.split("GPSTime_")[1].split("_C0")[0]): datum for datum in data_llo}
    
    data_lho = glob.glob(f"{dir}/H1/*LHO*FinalResults.txt")
    times_lho = {int(datum.split("GPSTime_")[1].split("_C0")[0]): datum for datum in data_lho}
        
    keys_llo = np.array(list(times_llo.keys())) 
    keys_lho = np.array(list(times_lho.keys())) 

    return {"H1": times_lho[keys_lho[np.argmin(np.abs(keys_lho - time))]], 
            "L1": times_llo[keys_llo[np.argmin(np.abs(keys_llo - time))]], 
            "V1": virgo}


def update(d, u, inplace=True):
    """Recursively update a dictionary."""
    if not inplace:
        d = deepcopy(d)
        u = deepcopy(u)
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



"""
This file contains code to allow unittests to be written with
Asimov so that productions can be tested with minimal boilerplate.
"""
from asimov.gitlab import find_events
from asimov import gitlab
from asimov import config
from asimov.cli import connect_gitlab


_, repository = connect_gitlab()

import unittest


class AsimovTest(unittest.TestCase):
    """
    Overloads the unittest.TestCase code.
    Simply makes `self.events` available to the test case.
    """
    @classmethod
    def setUpClass(cls):
        _, repository = connect_gitlab()
        cls.events = gitlab.find_events(repository)



"""RIFT Pipeline specification."""

import re
import time
import os
import shutil
import glob
import subprocess
from ..pipeline import Pipeline, PipelineException, PipelineLogger
from ..ini import RunConfiguration
from asimov import config, logger
from asimov.utils import set_directory
from ligo.gracedb.rest import HTTPError

class Rift(Pipeline):
    """
    The RIFT Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """
    name = "RIFT"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(Rift, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == "rift":
            raise PipelineException
        
        if "bootstrap" in self.production.meta:
            self.bootstrap = self.production.meta['bootstrap']
        else:
            self.bootstrap = False
        
    def after_completion(self):
        self.logger.info(f"Job has completed. Running PE Summary.",
                         production=self.production,
                         channels=['mattermost'])
        cluster = self.run_pesummary()
        self.production.meta['job id'] = int(cluster)
        self.production.status = "processing"

    def _convert_psd(self, ascii_format, ifo):
        """
        Convert an ascii format PSD to XML.

        Parameters
        ----------
        ascii_format : str
           The location of the ascii format file.
        ifo : str
           The IFO which this PSD is for.
        """
        command = ["convert_psd_ascii2xml",
                   "--fname-psd-ascii", f"{ascii_format}",
                   "--conventional-postfix",
                   "--ifo",  f"{ifo}"]
            
        pipe = subprocess.Popen(command, 
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        out, err = pipe.communicate()
        self.logger.info(command, production = self.production)
        if err:
            self.production.status = "stuck"
            if hasattr(self.production.event, "issue_object"):
                raise PipelineException(f"An XML format PSD could not be created.\n{command}\n{out}\n\n{err}",
                                            issue=self.production.event.issue_object,
                                            production=self.production.name)
            else:
                raise PipelineException(f"An XML format PSD could not be created.\n{command}\n{out}\n\n{err}",
                                        production=self.production.name)
        
            
    def before_submit(self, dryrun=False):
        """
        Convert the text-based PSD to an XML psd if the xml doesn't exist already.
        """
        event = self.production.event
        category = config.get("general", "calibration_directory")
        if len(self.production.get_psds("xml"))==0 and "psds" in self.production.meta:
            for ifo in self.production.meta['interferometers']:
                with set_directory(f"{event.work_dir}"):
                    sample = self.production.meta['quality']['sample-rate']
                    self._convert_psd(self.production.meta['psds'][sample][ifo], ifo, dryrun=dryrun)
                    asset = f"{ifo.upper()}-psd.xml.gz"
                    git_location = os.path.join(category, "psds")
                    self.production.event.repository.add_file(
                        asset,
                        os.path.join(git_location, str(sample), f"psd_{ifo}.xml.gz"),
                        commit_message = f"Added the xml format PSD for {ifo}.")

    def build_dag(self, user=None, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using util_RIFT_pseudo_pipe.py

        Parameters
        ----------
        user : str
           The user accounting tag which should be used to run the job.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.

        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.

        Notes
        -----

        In order to assemble the pipeline the RIFT runner requires additional
        production metadata: at least the l_max value.
        An example RIFT production specification would then look something like:
        
        ::
           
           - Prod3:
               rundir: tests/tmp/s000000xx/Prod3
               pipeline: rift
               approximant: IMRPhenomPv3
               lmax: 2
               cip jobs: 5 # This is optional, and will default to 3
               bootstrap: Prod1
               bootstrap fmin: 20
               needs: Prod1
               comment: RIFT production run.
               status: wait

        
        """
        cwd = os.getcwd()
        if self.production.event.repository:
            gps_file = self.production.get_timefile()
            try:
            
                coinc_file = self.production.get_coincfile()
                coinc_file = os.path.join(self.production.event.repository.directory, "C01_offline",
                                          coinc_file)
            except HTTPError:
                print("Unable to download the coinc file because it was not possible to connect to GraceDB")
                coinc_file = "COINC MISSING"

            try:
                
                ini = self.production.get_configuration().ini_loc
                ini = os.path.join(self.production.event.repository.directory, "C01_offline",  ini)
            except ValueError:
                print("Unable to find the configuration file. Have you run `$ asimov manage build` yet?")
                ini = "INI MISSING"
        else:
            gps_file = "gpstime.txt"
            ini = "INI MISSING"
            coinc_file = os.path.join(cwd, "coinc.xml")

        if self.production.get_meta("user"):
                user = self.production.get_meta("user")
        else:
            user = config.get("condor", "user")
            self.production.set_meta("user", user)

        os.environ['LIGO_USER_NAME'] = f"{user}"
        os.environ['LIGO_ACCOUNTING'] = f"{config.get('pipelines', 'accounting')}"

        try:
            calibration = config.get("general", "calibration")
        except:
            calibration = "C01"

        approximant = self.production.meta['approximant']


        if self.production.rundir:
            rundir = os.path.relpath(self.production.rundir, os.getcwd())
        else:
            rundir = os.path.join(os.path.expanduser("~"),
                                  self.production.event.name,
                                  self.production.name)
            self.production.rundir = rundir

        #lmax = self.production.meta['priors']['amp order']

        if "lmax" in self.production.meta:
            lmax = self.production.meta['likelihood']['lmax']
        elif "HM" in self.production.meta['approximant']:
            lmax = 4
        else:
            lmax = 2
            

        if "cip jobs" in self.production.meta['sampler']:
            cip = self.production.meta['cip jobs']
        else:
            cip = 3
            
        
        command = [os.path.join(config.get("pipelines", "environment"), "bin", "util_RIFT_pseudo_pipe.py"),
                   "--use-coinc", coinc_file,
                   "--l-max", f"{lmax}",
                   "--calibration", f"{calibration}",
                   "--add-extrinsic",
                   "--approx", f"{approximant}",
                   "--cip-explode-jobs", str(cip),
                   "--use-rundir", rundir,
                   "--ile-force-gpu",
                   "--use-ini", ini
        ]

        # If a starting frequency is specified, add it
        if "start-frequency" in self.production.meta:
            command += ["--fmin-template", self.production.quality['start-frequency']]

        # Placeholder LI grid bootstrapping; conditional on it existing and location specification
        
        if self.bootstrap:
            if self.bootstrap == "manual":
                if self.production.event.repository:
                    bootstrap_file = os.path.join(self.production.event.repository.directory, "C01_offline", f"{self.production.name}_bootstrap.xml.gz")
                else:
                    bootstrap_file = "{self.production.name}_bootstrap.xml.gz"
            else:
                raise PipelineException(f"Unable to find the bootstrapping production for {self.production.name}.",
                                            issue=self.production.event.issue_object,
                                            production=self.production.name)

            command += ["--manual-initial-grid", bootstrap_file]

        if dryrun:
            print(" ".join(command))

        else:
            self.logger.info(" ".join(command), production = self.production)

            with set_directory(self.production.event.work_dir):
                pipe = subprocess.Popen(command, 
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT)
                out, err = pipe.communicate()
                if err:
                    self.production.status = "stuck"
                    if hasattr(self.production.event, "issue_object"):
                        self.logger.info(out, production = self.production)
                        self.logger.error(err, production = self.production)
                        raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                                    issue=self.production.event.issue_object,
                                                    production=self.production.name)
                    else:
                        self.logger.info(out, production = self.production)
                        self.logger.error(err, production = self.production)
                        raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                                production=self.production.name)
                else:
                    if self.production.event.repository:
                        with set_directory(self.production.rundir):
                            for psdfile in self.production.get_psds("xml"):
                                ifo = psdfile.split("/")[-1].split("_")[1].split(".")[0]
                                os.system(f"cp {psdfile} {ifo}-psd.xml.gz")

                            #os.system("cat *_local.cache > local.cache")

                            if hasattr(self.production.event, "issue_object"):
                                return PipelineLogger(message=out,
                                                      issue=self.production.event.issue_object,
                                                      production=self.production.name)
                            else:
                                return PipelineLogger(message=out,
                                                      production=self.production.name)
    
    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster (using the RIFT dag name). This is an overwrite of the near identical parent function submit_dag()

        Parameters
        ----------
        category : str, optional
           The category of the job.
           Defaults to "C01_offline".
        production : str
           The production name.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """
        if not os.path.exists(self.production.rundir):
            os.makedirs(self.production.rundir)

        self.before_submit()
        with set_directory(self.production.rundir):
            if not dryrun:
                os.system("cat *_local.cache > local.cache")

            for psdfile in self.production.get_psds("xml"):
                ifo = psdfile.split("/")[-1].split("_")[1].split(".")[0]
                os.system(f"cp {psdfile} {ifo}-psd.xml.gz")

            command = ["condor_submit_dag", 
                       "-batch-name", f"rift/{self.production.event.name}/{self.production.name}",
                       os.path.join(self.production.rundir, "marginalize_intrinsic_parameters_BasicIterationWorkflow.dag")]

            if dryrun:
                print(" ".join(command))
            else:
                try:                    
                    dagman = subprocess.Popen(command,
                                              stdout=subprocess.PIPE,
                                              stderr=subprocess.STDOUT)
                    self.logger.info(command, production = self.production)
                except FileNotFoundError as error:
                    raise PipelineException("It looks like condor isn't installed on this system.\n"
                                            f"""I wanted to run {" ".join(command)}.""")

                stdout, stderr = dagman.communicate()


                if "submitted to cluster" in str(stdout):
                    cluster = re.search("submitted to cluster ([\d]+)", str(stdout)).groups()[0]
                    self.production.status = "running"
                    self.production.job_id = int(cluster)
                    return cluster, PipelineLogger(stdout)
                else:
                    raise PipelineException(f"The DAG file could not be submitted.\n\n{stdout}\n\n{stderr}",
                                            issue=self.production.event.issue_object,
                                            production=self.production.name)

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        try:
            count = self.production.meta['resurrections']
        except:
            count = 0
        count = len(glob.glob(os.path.join(self.production.rundir, "marginalize_intrinsic_parameters_BasicIterationWorkflow.dag.rescue*")))
        if "allow ressurect" in self.production.meta: count=0
        if (count < 90) and (len(glob.glob(os.path.join(self.production.rundir, "marginalize_intrinsic_parameters_BasicIterationWorkflow.dag.rescue*")))>0):
            count +=1
            #os.system("cat *_local.cache > local.cache")
            self.submit_dag()
        else:
            raise PipelineException("This job was resurrected too many times.",
                                    issue=self.production.event.issue_object,
                                    production=self.production.name)

    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and 
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/*.err") #+ glob.glob(f"{self.production.rundir}/*/logs/*")
        logs += glob.glob(f"{self.production.rundir}/*.out")
        messages = {}
        for log in logs:
            with open(log, "r") as log_f:
                message = log_f.read()
                messages[log.split("/")[-1]] = message
        return messages

    def detect_completion(self):
        """
        Check for the production of the posterior file to signal that the job has completed.
        """
        results_dir = glob.glob(f"{self.production.rundir}")
        if len(results_dir)>0: # dynesty_merge_result.json
            if len(glob.glob(os.path.join(results_dir[0], f"extrinsic_posterior_samples.dat"))) > 0:
                return True
            else:
                return False
        else:
            return False

    def samples(self):
        """
        Collect the combined samples file for PESummary.
        """
        return glob.glob(os.path.join(self.production.rundir, "extrinsic_posterior_samples.dat"))



"""Bilby Pipeline specification."""

import os
import glob
import subprocess

from asimov.utils import update

from ..pipeline import Pipeline, PipelineException, PipelineLogger
from ..ini import RunConfiguration
from .. import config, logger
import re
import numpy as np

from liquid import Liquid
import htcondor

class Bilby(Pipeline):
    """
    The Bilby Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """
    name = "Bilby"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(Bilby, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == "bilby":
            raise PipelineException

    def detect_completion(self):
        """
        Check for the production of the posterior file to signal that the job has completed.
        """
        results_dir = glob.glob(f"{self.production.rundir}/result")
        if len(results_dir)>0: # dynesty_merge_result.json
            if len(glob.glob(os.path.join(results_dir[0], f"*merge*_result.*"))) > 0:
                return True
            else:
                return False
        else:
            return False


    def _determine_prior(self):
        """
        Determine the correct choice of prior file for this production.
        """

        if "prior file" in self.production.meta:
            return self.production.meta['prior file']
        else:
            duration = int(self.production.meta['data']['segment length'])
            scale_factor = 1 # Check this
            template = None

            prior_parameters = {}
            
            if "event type" in self.production.meta:
                event_type = self.production.meta['event type'].lower()
            else:
                event_type = "bbh"
                self.production.meta['event type'] = event_type

                if self.production.event.issue_object:
                    self.production.event.issue_object.update_data()

            if template is None:
                template_filename = f"{event_type}.prior.template"
                try:
                    template = os.path.join(config.get("bilby", "priors"), template_filename)
                except:
                    from pkg_resources import resource_filename
                    template = resource_filename("asimov", f'priors/{template_filename}')

            #with open(template, "r") as old_prior:
            #    prior_string = old_prior.read().format(**prior_parameters)

            priors = {}
            priors = update(priors, self.production.event.ledger.data['priors'])
            priors = update(priors, self.production.event.meta['priors'])
            priors = update(priors, self.production.meta['priors'])
            
            liq = Liquid(template)
            rendered = liq.render(priors=priors, config=config)
            
            prior_name = f"{self.production.name}.prior"
            prior_file = os.path.join(os.getcwd(), prior_name)
            with open(prior_file, "w") as new_prior:
                new_prior.write(rendered)
                
            repo = self.production.event.repository
            try:
                repo.add_file(prior_file, os.path.join("C01_offline", prior_name))
                os.remove(prior_file)
            except:
                pass
            return os.path.join(self.production.event.repository.directory, "C01_offline", prior_name)

        
    def build_dag(self, psds=None, user=None, clobber_psd=False, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using bilby_pipe.

        Parameters
        ----------
        production : str
           The production name.
        psds : dict, optional
           The PSDs which should be used for this DAG. If no PSDs are
           provided the PSD files specified in the ini file will be used
           instead.
        user : str
           The user accounting tag which should be used to run the job.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.

        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """

        cwd = os.getcwd()

        self.logger.info(f"[bilby] Working in {cwd}")

        if self.production.event.repository:
            ini = self.production.event.repository.find_prods(
                self.production.name,
                self.category)[0]
            ini = os.path.join(cwd, ini)
            gps_file = self.production.get_timefile()
        else:
            gps_file = "gpstime.txt"
            ini = f"{self.production.name}.ini"
            
        if self.production.rundir:
            rundir = self.production.rundir
        else:
            rundir = os.path.join(os.path.expanduser("~"),
                                  self.production.event.name,
                                  self.production.name)
            self.production.rundir = rundir

        if "job label" in self.production.meta:
            job_label = self.production.meta['job label']
        else:
            job_label = self.production.name

        prior_file = self._determine_prior()
        
        command = [os.path.join(config.get("pipelines", "environment"), "bin", "bilby_pipe"),
                   ini,
                   "--label", job_label,
                   "--outdir", f"{cwd}/{self.production.rundir}",
                   "--accounting", f"{self.production.meta['scheduler']['accounting group']}"
        ]

        if dryrun:
            print(" ".join(command))
        else:
            self.logger.info(" ".join(command))
            pipe = subprocess.Popen(command, 
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            out, err = pipe.communicate()
            self.logger.info(out)
            self.logger.error(err)
            #os.chdir(cwd)
            if err or "DAG generation complete, to submit jobs" not in str(out):
                self.production.status = "stuck"
                if hasattr(self.production.event, "issue_object"):
                    raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                                issue=self.production.event.issue_object,
                                                production=self.production.name)
                else:
                    raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                            production=self.production.name)
            else:
                if hasattr(self.production.event, "issue_object"):
                    return PipelineLogger(message=out,
                                          issue=self.production.event.issue_object,
                                          production=self.production.name)
                else:
                    return PipelineLogger(message=out,
                                          production=self.production.name)


    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        dryrun : bool
           If set to true the DAG will not be submitted, but all commands will be printed to standard output instead. Defaults to False.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.

        Notes
        -----
        This overloads the default submission routine, as bilby seems to store
        its DAG files in a different location
        """
        #os.chdir(self.production.event.meta['working directory'])   
        #os.chdir(os.path.join(self.production.event.repository.directory,
        #                      self.category))


        self.before_submit()
        
        try:
            # to do: Check that this is the correct name of the output DAG file for billby (it
            # probably isn't)
            if "job label" in self.production.meta:
                job_label = self.production.meta['job label']
            else:
                job_label = self.production.name
            dag_filename = f"dag_{job_label}.submit"
            command = [
                #"ssh", f"{config.get('scheduler', 'server')}",
                "condor_submit_dag",
                       "-batch-name", f"bilby/{self.production.event.name}/{self.production.name}",
                                   os.path.join(self.production.rundir, "submit", dag_filename)]

            if dryrun:
                print(" ".join(command))
            else:
                dagman = subprocess.Popen(command,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)

                stdout, stderr = dagman.communicate()

                if "submitted to cluster" in str(stdout):
                    cluster = re.search("submitted to cluster ([\d]+)", str(stdout)).groups()[0]
                    self.production.status = "running"
                    self.production.job_id = int(cluster)
                    return cluster, PipelineLogger(stdout)
                else:
                    raise PipelineException(f"The DAG file could not be submitted.\n\n{stdout}\n\n{stderr}",
                                            issue=self.production.event.issue_object,
                                            production=self.production.name)
        except FileNotFoundError as error:
            raise PipelineException("It looks like condor isn't installed on this system.\n"
                                    f"""I wanted to run {" ".join(command)}.""")

        
    def collect_assets(self):
        """
        Gather all of the results assets for this job.
        """
        return {"samples": self.samples()}

    def samples(self):
        """
        Collect the combined samples file for PESummary.
        """
        return glob.glob(os.path.join(self.production.rundir, "result", "*_merge*_result.*"))
        
    def after_completion(self):
        self.logger.info(f"Job has completed. Running PE Summary.", production=self.production, channels=['mattermost'])
        cluster = self.run_pesummary()
        self.production.meta['job id'] = int(cluster)
        self.production.status = "processing"
        self.production.event.update_data()

    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and 
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/submit/*.err") + glob.glob(f"{self.production.rundir}/log*/*.err")
        logs += glob.glob(f"{self.production.rundir}/*/*.out")
        messages = {}
        for log in logs:
            try:
                with open(log, "r") as log_f:
                    message = log_f.read()
                    message = message.split("\n")
                    messages[log.split("/")[-1]] = "\n".join(message[-100:])
            except:
                messages[log.split("/")[-1]] = "There was a problem opening this log file."
        return messages

    def check_progress(self):
        """
        Check the convergence progress of a job.
        """
        logs = glob.glob(f"{self.production.rundir}/log_data_analysis/*.out")
        messages = {}
        for log in logs:
            try:
                with open(log, "r") as log_f:
                    message = log_f.read()
                    message = message.split("\n")[-1]
                    p = re.compile(r"([\d]+)it")
                    iterations = p.search(message)
                    p = re.compile(r"dlogz:([\d]*\.[\d]*)")
                    dlogz = p.search(message)
                    if iterations:
                        messages[log.split("/")[-1]] = (iterations.group(), dlogz.group())
            except:
                messages[log.split("/")[-1]] = "There was a problem opening this log file."
        return messages

    @classmethod
    def read_ini(cls, filepath):
        """ 
        Read and parse a bilby configuration file.

        Note that bilby configurations are property files and not compliant ini configs.

        Parameters
        ----------
        filepath: str
           The path to the ini file.
        """

        with open(filepath, "r") as f:
            file_content = '[root]\n' + f.read()

        config_parser = ConfigParser.RawConfigParser()
        config_parser.read_string(file_content)

        return config_parser

    def html(self):
        """Return the HTML representation of this pipeline."""
        pages_dir = os.path.join(self.production.event.name, self.production.name)
        out = ""

        return out
    
    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        try:
            count = self.production.meta['resurrections']
        except:
            count = 0
        if (count < 5) and (len(glob.glob(os.path.join(self.production.rundir, "submit", "*.rescue*")))>0):
            count +=1
            self.submit_dag()



import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from asimov.pipelines.bayeswave import BayesWave
from asimov.pipelines.lalinference import LALInference
from asimov.pipelines.rift import Rift
from asimov.pipelines.bilby import Bilby
    
discovered_pipelines = entry_points(group='asimov.pipelines')


known_pipelines = {"bayeswave": BayesWave,
                   "bilby": Bilby,
                   "rift": Rift,
                   "lalinference": LALInference}


for pipeline in discovered_pipelines:
    known_pipelines[pipeline.name] = pipeline.load()



"""
Reporting functions
"""
import yaml
import glob
from datetime import datetime
import pytz
import os

import click

tz = pytz.timezone('Europe/London')


import otter
import otter.bootstrap as bt

from asimov import current_ledger
from asimov.cli import known_pipelines
from asimov import config

@click.group()
def report():
    """Produce reports about the state of the project.
    """
    pass

@click.option("--location", "webdir", default=None, help="The place to save the report to")
@click.argument("event", default=None, required=False)
@report.command()
def html(event, webdir):
    """
    Return the ledger for a given event.
    If no event is specified then the entire production ledger is returned.
    """

    events = current_ledger.get_event(event)

    if not webdir:
        webdir = config.get("general", "webroot")

    report = otter.Otter(f"{webdir}/index.html", 
                         author="Asimov", 
                         title="Asimov project report", 
                         #author_email=config.get("report", "report_email"),
                         config_file="asimov.conf"
    )
    with report:

        status_map = {"cancelled": "light",
                      "finished": "success",
                      "uploaded": "success",
                      "processing": "primary",
                      "running": "primary",
                      "stuck": "warning",
                      "restart": "secondary",
                      "ready": "secondary",
                      "wait": "light",
                      "stop": "danger",
                      "manual": "light",
                      "stopped": "light"}

        review_map = {"deprecated": "warning",
                      "none": "default",
                      "approved": "success",
                      "rejected": "danger",
                      "checked": "info"
                      }

        style = """
<style>
        body {
        background-color: #f2f2f2;
        }
        
        .review-deprecated, .status-cancelled, .review-rejected {
        display: none;
        }

        .event-data {
        margin: 1rem;
        margin-bottom: 2rem;
        }

        .asimov-sidebar {
        position: sticky;
        top: 4rem;
        height: calc(100vh - 4rem);
        overflow-y: auto;
        }

        .asimov-analysis {
        padding: 1rem;
        background: lavenderblush;
        margin: 0.5rem;
        border-radius: 0.5rem;
        }

        .asimov-analysis-running, .asimov-analysis-processing {
        background: #DEEEFF;
        }
        
        .asimov-analysis-finished, .asimov-analysis-uploaded {
        background: #E1EDE4;
        }
        .asimov-analysis-stuck {
        background: #FFF5D9;
        }
        
        .asimov-status {
        float: right;
        clear: both;
        }
</style>
        """
        report + style

    with report:
        navbar = bt.Navbar(f"Asimov  |  {current_ledger.data['project']['name']}", background="navbar-dark bg-primary")
        report + navbar

    events = sorted(events, key = lambda a: a.name)
    cards = "<div class='container-fluid'><div class='row'><div class='col-12 col-md-3 col-xl-2  asimov-sidebar'>"
    
    toc = """<nav><h6>Subjects</h6><ul class="list-unstyled">"""
    for event in events:
        toc += f"""<li><a href="#card-{event.name}">{event.name}</a></li>"""
        
    toc += "</ul></nav>"

    cards += toc
    cards += """</div><div class='events col-md-9 col-xl-10' data-isotope='{ "itemSelector": ".production-item", "layoutMode": "fitRows" }'>"""


    for event in events:
        card = ""
        # This is a quick test to try and improve readability
        card += event.html()
        
        # card += """<p class="card-text">Card text</p>""" # 
        card += """
</div>
</div>"""
        cards += card

    cards += "</div></div>"
    with report:
        report + cards

    with report:
        time = f"Report generated at {datetime.now(tz):%Y-%m-%d %H:%M}"
        report + time
        

@click.argument("event", default=None, required=False)
@report.command()
def status(event):
    """
    Provide a simple summary of the status of a given event.

    Arguments
    ---------
    name : str, optional
       The name of the event.

    """
    for event in current_ledger.get_event(event):
        click.secho(f"{event.name:30}", bold=True)
        if len(event.productions)>0:
            click.secho("\tProductions", bold=True)
            for production in event.productions:
                click.echo(f"\t- {production.name} " + click.style(f"{production.pipeline}") + " " + click.style(f"{production.status}"))
        if len(event.get_all_latest())>0:
            click.secho("\tProductions waiting: ", bold=True, nl=False)
            waiting = event.get_all_latest()
            for awaiting in waiting:
                click.echo(f"{awaiting.name} ", nl=False)
            click.echo("")

@click.option("--yaml", "yaml_f", default=None, help="A YAML file to save the ledger to.")
@click.argument("event", default=None, required=False)
@report.command()
def ledger(event, yaml_f):
    """
    Return the ledger for a given event.
    If no event is specified then the entire production ledger is returned.
    """
    total = []
    for event in current_ledger.get_event(event):
        total.append(yaml.safe_load(event.to_yaml()))

    click.echo(yaml.dump(total))

    if yaml_f:
        with open(yaml_f, "w") as f:
            f.write(yaml.dump(total))



"""
Job endpoints
=============

This module provides endpoints for interaction with job schedulers and individual analysis jobs.

"""

from flask import Blueprint, request, jsonify, make_response
import asimov
from asimov.condor import CondorJobList

bp = Blueprint("asimov", __name__, url_prefix="jobs")

@bp.route("/", methods=['GET'])
def job_list():
    """
    Get the list of all running jobs for this user.

    Example
    -------

    .. http:example:: curl wget python-requests

       GET /jobs/ HTTP/1.1
       Content-Type: application/json


       HTTP/1.1 200 OK
       Content-Type: application/json

       "asimov"

    """

    return make_response(jsonify(CondorJobList()), 200)



"""
Endpoints for monitoring asimov jobs.

Asimov exposes a number of end-point which can be used to
get a rapid overview of the run status of a set of analyses
it is managing.
"""
from flask import Blueprint, request, jsonify, make_response, json
from asimov import ledger

monitor_bp = Blueprint("monitor", __name__, url_prefix="/monitor")

@monitor_bp.route("/", methods=["GET"])
@monitor_bp.route("/<event>/", methods=["GET"])
@monitor_bp.route("/<event>/<production>", methods=["GET"])
def show_production_status(event=None, production=None):
    """Show the run status of either a given production, 
    all the productions for an event, or all of the productions
    in the project.

    .. http::get:: /monitor/
    
       Monitor all productions.

       .. http:example:: curl wget python-requests

          GET /monitor/ HTTP/1.1
          Content-Type: application/json
    """
    return make_response(jsonify(None), 200)




import shutil

import click

from asimov.cli import ACTIVE_STATES, known_pipelines
from asimov import gitlab
from asimov import config
from asimov import current_ledger as ledger
from asimov import logger
from asimov import condor

from asimov.cli import manage
from asimov.cli import report

@click.option("--dry-run", "-n", "dry_run", is_flag=True)
@click.command()
def start(dry_run):
    """Set up a cron job on condor to monitor the project."""

    submit_description = {
          "executable": shutil.which("asimov"),  
          "arguments": "monitor --chain",
          "accounting_group": config.get("pipelines", "accounting"),
          "output": f"asimov_cron.out",
          "on_exit_remove": "false",
          "error": f"asimov_cron.err",
          "log": f"asimov_cron.log",
          "request_cpus": "1",
          "cron_minute": "*/15",
          "getenv": "true",
          "batch_name": f"asimov/monitor/{ledger.data['project']['name']}",
          "request_memory": "8192MB",
          "request_disk": "8192MB",
    }
    cluster = condor.submit_job(submit_description)
    ledger.data['cronjob'] = cluster
    ledger.save()
    click.secho(f"  \t  ● Asimov is running ({cluster})", fg="green")
    
@click.option("--dry-run", "-n", "dry_run", is_flag=True)
@click.command()
def stop(dry_run):
    """Set up a cron job on condor to monitor the project."""
    cluster = ledger.data['cronjob']
    condor.delete_job(cluster)
    click.secho(f"  \t  ● Asimov has been stopped", fg="red")
    
@click.argument("event", default=None, required=False)
@click.option("--update", "update", default=False,
              help="Force the git repos to be pulled before submission occurs.")
@click.option("--dry-run", "-n", "dry_run", is_flag=True)
@click.option("--chain", "-c", "chain", default=False,
              is_flag=True,
              help="Chain multiple asimov commands")
@click.command()
@click.pass_context
def monitor(ctx, event, update, dry_run, chain):
    """
    Monitor condor jobs' status, and collect logging information.
    """

    if chain:
        ctx.invoke(manage.build, event=event)
        ctx.invoke(manage.submit, event=event)
    
    # First pull the condor job listing
    job_list = condor.CondorJobList()
    
    for event in ledger.get_event(event):
        stuck = 0
        running = 0
        ready = 0
        finish = 0
        click.secho(f"{event.name}", bold=True)
        on_deck = [production
                   for production in event.productions
                   if production.status.lower() in ACTIVE_STATES]
        for production in on_deck:

            click.echo(f"\t- " + click.style(f"{production.name}", bold=True) + click.style(f"[{production.pipeline}]", fg = "green"))

            # Jobs marked as ready can just be ignored as they've not been stood-up
            if production.status.lower() == "ready":
                click.secho(f"  \t  ● {production.status.lower()}", fg="green")
                continue
            
            # Deal with jobs which need to be stopped first
            if production.status.lower() == "stop":
                pipe = production.pipeline
                if not dry_run:
                    pipe.eject_job()
                    production.status = "stopped"
                    click.secho(f"  \tStopped", fg="red")
                else:
                    click.echo("\t\t{production.name} --> stopped")
                continue

            # Get the condor jobs
            try:
                if "job id" in production.meta:
                    if not dry_run:
                        if production.meta['job id'] in job_list.jobs:
                            job = job_list.jobs[production.meta['job id']]
                        else:
                            job = None
                    else:
                        click.echo(f"\t\tRunning under condor")
                else:
                    raise ValueError # Pass to the exception handler

                if not dry_run:
                    if job.status.lower() == "running" and production.status == "processing":
                        click.echo(f"  \t  " + click.style("●", "green") + f" Postprocessing for {production.name} is running (condor id: {production.meta['job id']})")
                    elif job.status.lower() == "running":
                        click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} is running (condor id: {production.meta['job id']})")
                        if not "profiling" in production.meta: production.meta['profiling'] = {}
                        #production.meta['profiling'] = job.get_data()['WallClockCheckpoint']
                        production.status = "running"
                    elif job.status.lower() == "completed":
                        click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} has completed")
                    #elif job.status.lower() == "processing":
                    #    click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} is postprocessing (condor id: {production.meta['job id']})")
                    elif job.status.lower() == "held":
                        click.echo(f"  \t  " + click.style("●", "yellow") + f" {production.name} is held on the scheduler (condor id: {production.meta['job id']})")
                        production.status = "stuck"
                        stuck += 1
                    else:
                        running += 1

            except (ValueError, AttributeError) as e:
                if production.pipeline:
                    
                    pipe = production.pipeline

                    if production.status.lower() == "stop":
                        pipe.eject_job()
                        production.status = "stopped"
                        click.echo(f"  \t  " + click.style("●", "red") + f" {production.name} has been stopped")
                        job_list.refresh()                        
                    elif production.status.lower() == "finished":
                        pipe.after_completion()
                        click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} has finished and post-processing has been started")
                        job_list.refresh()
                    elif production.status.lower() == "processing":
                    # Need to check the upload has completed
                        if pipe.detect_completion_processing():
                            try:
                                pipe.after_processing()
                                click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} has been finalised and stored")
                            except ValueError as e:
                                click.echo(e)
                        else:
                            click.echo(f"  \t  " + click.style("●", "green") + f" {production.name} has finished and post-processing is running")
                    elif pipe.detect_completion() and production.status.lower() == "running":
                        # The job has been completed, collect its assets
                        if not "profiling" in production.meta: production.meta['profiling'] = {}
                        try:
                            production.meta['profiling'] = condor.collect_history(production.meta['job id'])
                            production.meta['job id'] = None
                        except condor.htcondor.HTCondorIOError:
                            #self.logger.error("Could not collect condor profiling data.")
                            pass

                        finish += 1
                        production.status = "finished"
                        pipe.after_completion()
                        click.secho(f"  \t  ● {production.name} - Completion detected", fg="green")
                        job_list.refresh()
                    else:
                        # It looks like the job has been evicted from the cluster
                        click.echo(f"  \t  " + click.style("●", "yellow") + f" {production.name} is stuck; attempting a rescue")
                        try:
                            pipe.resurrect()
                        except:
                            production.status = "stuck"
                            click.echo(f"  \t  " + click.style("●", "red") + f" {production.name} is stuck; automatic rescue was not possible")

                if production.status == "stuck":
                    click.echo(f"  \t  " + click.style("●", "yellow") + f" {production.name} is stuck")
                    
            ledger.update_event(event)

            if chain:
                ctx.invoke(report.html)



"""
The asimov server
-----------------
This feature allows asimov to monitor and interact with ongoing
analyses without the need for an ongoing cronjob.
"""
from flask import Flask


from flask.json import JSONEncoder
from datetime import date
from . import logging
from . import basic

import asimov.server.ledger
import asimov.server.productions
import asimov.server.monitor


from flask.json import JSONEncoder
from datetime import date


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        try:
            if isinstance(obj, date):
                return obj.isoformat()
            iterable = iter(obj)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, obj)


def create_app():
    """
    Create the web app for asimov
    """
    app = Flask("asimov")
    
    app.json_encoder = CustomJSONEncoder

    app.register_blueprint(logging.bp)
    app.register_blueprint(basic.bp)
    
    app.register_blueprint(asimov.server.ledger.events_bp)
    app.register_blueprint(asimov.server.monitor.monitor_bp)
    app.register_blueprint(asimov.server.productions.prods_bp)

    app.json_encoder = CustomJSONEncoder
    return app


def bootstrap():
    app = create_app()
    app.run(debug=False)


if __name__ == '__main__':
    bootstrap()



"""
Tools for adding data from JSON and YAML files.
Inspired by the kubectl apply approach from kubernetes.
"""

import yaml

import click
import requests

from asimov import current_ledger as ledger
from asimov.utils import update
import asimov.event


def apply_page(file, event, ledger=ledger):
    if file[:4] == "http":
        r = requests.get(file)
        if r.status_code == 200:
            data = r.text
        else:
            raise ValueError(f"Could not download this file: {file}")
    else:
        with open(file, "r") as apply_file:
            data = apply_file.read()

    quick_parse = yaml.safe_load_all(data) # Load as a dictionary so we can identify the object type it contains

    for document in quick_parse:e

        if document['kind'] == "event":
            document.pop("kind")
            event = asimov.event.Event.from_yaml(yaml.dump(document))
            ledger.update_event(event)
            click.echo(click.style("●", fg="green") + f" Successfully applied {event.name}")

        elif document['kind'] == "analysis":
            document.pop("kind")
            if event:
                event_s = event
            else:
                if "event" in document:
                    event_s = document['event']
                else:
                    prompt = "Which event should these be applied to?"
                    event_s = str(click.prompt(prompt))
            try:    
                event_o = ledger.get_event(event_s)[0]
            except KeyError:
                click.echo(click.style("●", fg="red") + f" Could not apply a production, couldn't find the event {event}")
                continue
            production = asimov.event.Production.from_dict(document, event=event_o)
            try:
                event_o.add_production(production)
                ledger.update_event(event_o)
                click.echo(click.style("●", fg="green") + f" Successfully applied {production.name} to {event_o.name}")
            except ValueError:
                click.echo(click.style("●", fg="red") + f" Could not apply {production.name} to {event_o.name} as an analysis already exists with this name")

        elif document['kind'] == "configuration":
            document.pop("kind")
            update(ledger.data, document)
            ledger.save()
            click.echo(click.style("●", fg="green") + f" Successfully applied a configuration update")


@click.command()
@click.option("--file", "-f", 
              help="Location of the file containing the ledger items.")
@click.option("--event", "-e",
              help="The event which the ledger items should be applied to (e.g. for analyses)",
              default=None)
def apply(file, event):
    apply_page(file, event)



import requests
import json
from . import config

class Mattermost(object):

    def __init__(self, url=None):
        if not url:
            self.url = config.get("mattermost", "webhook_url")
        else:
            self.url = url

    def send_message(self, message, channel=None):
        """
        Send a message to a chat channel.
        
        Parameters
        ----------
        message : str 
           The text of the message.
        channel : str, optional
           The name of the channel.
           To send a direct message to a person prefix their username with
           an @ sign. Defaults to the default channel set in mattermost
           for the webhook.
        """
        self.submit_payload(message, channel)
            
    def submit_payload(self, message, channel=None):
        """
        Send a payload (normally a message) to a chat channel.

        Parameters
        ----------
        message : str 
           The text of the message.
        channel : str, optional
           The name of the channel.
           To send a direct message to a person prefix their username with
           an @ sign. Defaults to the default channel set in mattermost
           for the webhook.
        """
        data = {"text": message}

        if channel:
            data['channel'] = channel
        
        response = requests.post(self.url, data=json.dumps(data),
                                 headers={"Content-Type": "application/json"})

        if response.status_code != 200:
            raise ValueError("Mattermost post failed")
